@@ -15,6 +15,7 @@
 	Authors: 	Domas Mituzas, Facebook ( domas at fb dot com )
 			Mark Leith, Oracle Corporation (mark dot leith at oracle dot com)
 			Andrew Hutchings, SkySQL (andrew at skysql dot com)
+			Max Bubenick, Percona RDBA (max dot bubenick at percona dot com)
 */
 
 #define _LARGEFILE64_SOURCE
@@ -38,6 +39,7 @@
 #include "common.h"
 #include "g_unix_signal.h"
 #include "config.h"
+#include <math.h>
 
 char *regexstring=NULL;
 
@@ -51,9 +53,10 @@ static GMutex * init_mutex = NULL;
 gchar *output_directory= NULL;
 guint statement_size= 1000000;
 guint rows_per_file= 0;
+guint chunk_filesize = 0;
 int longquery= 60;
 int build_empty_files= 0;
-
+int skip_tz= 0;
 int need_dummy_read= 0;
 int compress_output= 0;
 int killqueries= 0;
@@ -76,12 +79,16 @@ FILE *logoutfile= NULL;
 
 gboolean no_schemas= FALSE;
 gboolean no_locks= FALSE;
+gboolean less_locking = FALSE;
+gboolean use_savepoints = FALSE;
+gboolean success_on_1146 = FALSE;
 
 GList *innodb_tables= NULL;
 GList *non_innodb_table= NULL;
 GList *table_schemas= NULL;
 gint non_innodb_table_counter= 0;
 gint non_innodb_done= 0;
+guint less_locking_threads = 0;
 
 // For daemon mode, 0 or 1
 guint dump_number= 0;
@@ -89,6 +96,8 @@ guint binlog_connect_id= 0;
 gboolean shutdown_triggered= FALSE;
 GAsyncQueue *start_scheduled_dump;
 GMainLoop *m1;
+static GCond * ll_cond = NULL; 
+static GMutex * ll_mutex = NULL;
 
 int errors;
 
@@ -98,19 +107,25 @@ static GOptionEntry entries[] =
 	{ "tables-list", 'T', 0, G_OPTION_ARG_STRING, &tables_list, "Comma delimited table list to dump (does not exclude regex option)", NULL },
 	{ "outputdir", 'o', 0, G_OPTION_ARG_FILENAME, &output_directory, "Directory to output files to",  NULL },
 	{ "statement-size", 's', 0, G_OPTION_ARG_INT, &statement_size, "Attempted size of INSERT statement in bytes, default 1000000", NULL},
-	{ "rows", 'r', 0, G_OPTION_ARG_INT, &rows_per_file, "Try to split tables into chunks of this many rows", NULL},
+	{ "rows", 'r', 0, G_OPTION_ARG_INT, &rows_per_file, "Try to split tables into chunks of this many rows. This option turns off --chunk-filesize", NULL},
+	{ "chunk-filesize", 'F', 0, G_OPTION_ARG_INT, &chunk_filesize, "Split tables into chunks of this output file size. This value is in MB", NULL },
 	{ "compress", 'c', 0, G_OPTION_ARG_NONE, &compress_output, "Compress output files", NULL},
 	{ "build-empty-files", 'e', 0, G_OPTION_ARG_NONE, &build_empty_files, "Build dump files even if no data available from table", NULL},
 	{ "regex", 'x', 0, G_OPTION_ARG_STRING, &regexstring, "Regular expression for 'db.table' matching", NULL},
 	{ "ignore-engines", 'i', 0, G_OPTION_ARG_STRING, &ignore_engines, "Comma delimited list of storage engines to ignore", NULL },
 	{ "no-schemas", 'm', 0, G_OPTION_ARG_NONE, &no_schemas, "Do not dump table schemas with the data", NULL },
 	{ "no-locks", 'k', 0, G_OPTION_ARG_NONE, &no_locks, "Do not execute the temporary shared read lock.  WARNING: This will cause inconsistent backups", NULL },
+	{ "less-locking", 0, 0, G_OPTION_ARG_NONE, &less_locking, "Minimize locking time on InnoDB tables.", NULL},
 	{ "long-query-guard", 'l', 0, G_OPTION_ARG_INT, &longquery, "Set long query timer in seconds, default 60", NULL },
 	{ "kill-long-queries", 'k', 0, G_OPTION_ARG_NONE, &killqueries, "Kill long running queries (instead of aborting)", NULL },
 	{ "binlogs", 'b', 0, G_OPTION_ARG_NONE, &need_binlogs, "Get a snapshot of the binary logs as well as dump data",  NULL },
 	{ "daemon", 'D', 0, G_OPTION_ARG_NONE, &daemon_mode, "Enable daemon mode", NULL },
 	{ "snapshot-interval", 'I', 0, G_OPTION_ARG_INT, &snapshot_interval, "Interval between each dump snapshot (in minutes), requires --daemon, default 60", NULL },
 	{ "logfile", 'L', 0, G_OPTION_ARG_FILENAME, &logfile, "Log file name to use, by default stdout is used", NULL },
+	{ "tz-utc", 0, 0, G_OPTION_ARG_NONE, NULL, "SET TIME_ZONE='+00:00' at top of dump to allow dumping of TIMESTAMP data when a server has data in different time zones or data is being moved between servers with different time zones, defaults to on use --skip-tz-utc to disable.", NULL },
+	{ "skip-tz-utc", 0, 0, G_OPTION_ARG_NONE, &skip_tz, "", NULL },
+	{ "use-savepoints", 0, 0, G_OPTION_ARG_NONE, &use_savepoints, "Use savepoints to reduce metadata locking issues, needs SUPER privilege", NULL },
+	{ "success-on-1146", 0, 0, G_OPTION_ARG_NONE, &success_on_1146, "Not increment error count and Warning instead of Critical in case of table doesn't exist", NULL},
 	{ NULL, 0, 0, G_OPTION_ARG_NONE,   NULL, NULL, NULL }
 };
 
@@ -119,7 +134,8 @@ struct tm tval;
 void dump_schema_data(MYSQL *conn, char *database, char *table, char *filename);
 void dump_schema(char *database, char *table, struct configuration *conf);
 void dump_table(MYSQL *conn, char *database, char *table, struct configuration *conf, gboolean is_innodb);
-guint64 dump_table_data(MYSQL *, FILE *, char *, char *, char *);
+void dump_tables(MYSQL *, GList *, struct configuration *);
+guint64 dump_table_data(MYSQL *, FILE *, char *, char *, char *, char *);
 void dump_database(MYSQL *, char *);
 GList * get_chunks_for_table(MYSQL *, char *, char*,  struct configuration *conf);
 guint64 estimate_count(MYSQL *conn, char *database, char *table, char *field, char *from, char *to);
@@ -307,7 +323,7 @@ void *process_queue(struct thread_data *td) {
 	g_mutex_lock(init_mutex);
 	MYSQL *thrconn = mysql_init(NULL);
 	g_mutex_unlock(init_mutex);
-
+	
 	mysql_options(thrconn,MYSQL_READ_DEFAULT_GROUP,"mydumper");
 
 	if (compress_protocol)
@@ -319,7 +335,11 @@ void *process_queue(struct thread_data *td) {
 	} else {
 		g_message("Thread %d connected using MySQL connection ID %lu", td->thread_id, mysql_thread_id(thrconn));
 	}
-
+	
+	if(use_savepoints && mysql_query(thrconn, "SET SQL_LOG_BIN = 0")){
+		g_critical("Failed to disable binlog for the thread: %s",mysql_error(thrconn));
+		exit(EXIT_FAILURE);
+	}
 	if ((detected_server == SERVER_TYPE_MYSQL) && mysql_query(thrconn, "SET SESSION wait_timeout = 2147483")){
 		g_warning("Failed to increase wait_timeout: %s", mysql_error(thrconn));
 	}
@@ -330,6 +350,10 @@ void *process_queue(struct thread_data *td) {
 		g_critical("Failed to start consistent snapshot: %s",mysql_error(thrconn));
 		errors++;
 	}
+	if(!skip_tz && mysql_query(thrconn, "/*!40103 SET TIME_ZONE='+00:00' */")){
+		g_critical("Failed to set time zone: %s",mysql_error(thrconn));
+	}
+
 	/* Unfortunately version before 4.1.8 did not support consistent snapshot transaction starts, so we cheat */
 	if (need_dummy_read) {
 		mysql_query(thrconn,"SELECT /*!40001 SQL_NO_CACHE */ * FROM mysql.mydumperdummy");
@@ -345,12 +369,25 @@ void *process_queue(struct thread_data *td) {
 	struct table_job* tj= NULL;
 	struct schema_job* sj= NULL;
 	struct binlog_job* bj= NULL;
+	
+	/* if less locking we need to wait until that threads finish
+	    progressively waking up this threads */
+	if(less_locking){
+		g_mutex_lock(ll_mutex);
+		
+		while (less_locking_threads >= td->thread_id) {
+			g_cond_wait (ll_cond, ll_mutex);
+		}
+		
+		g_mutex_unlock(ll_mutex);
+	}
+		
 	for(;;) {
+		
 		GTimeVal tv;
 		g_get_current_time(&tv);
 		g_time_val_add(&tv,1000*1000*1);
 		job=(struct job *)g_async_queue_pop(conf->queue);
-
 		if (shutdown_triggered && (job->type != JOB_SHUTDOWN)) {
 			continue;
 		}
@@ -362,7 +399,13 @@ void *process_queue(struct thread_data *td) {
 					g_message("Thread %d dumping data for `%s`.`%s` where %s", td->thread_id, tj->database, tj->table, tj->where);
 				else
 					g_message("Thread %d dumping data for `%s`.`%s`", td->thread_id, tj->database, tj->table);
+				if(use_savepoints && mysql_query(thrconn, "SAVEPOINT mydumper")){
+					g_critical("Savepoint failed: %s",mysql_error(thrconn));
+				}
 				dump_table_data_file(thrconn, tj->database, tj->table, tj->where, tj->filename);
+				if(use_savepoints && mysql_query(thrconn, "ROLLBACK TO SAVEPOINT mydumper")){
+					g_critical("Rollback to savepoint failed: %s",mysql_error(thrconn));
+				}
 				if(tj->database) g_free(tj->database);
 				if(tj->table) g_free(tj->table);
 				if(tj->where) g_free(tj->where);
@@ -376,7 +419,13 @@ void *process_queue(struct thread_data *td) {
 					g_message("Thread %d dumping data for `%s`.`%s` where %s", td->thread_id, tj->database, tj->table, tj->where);
 				else
 					g_message("Thread %d dumping data for `%s`.`%s`", td->thread_id, tj->database, tj->table);
+				if(use_savepoints && mysql_query(thrconn, "SAVEPOINT mydumper")){
+					g_critical("Savepoint failed: %s",mysql_error(thrconn));
+				}
 				dump_table_data_file(thrconn, tj->database, tj->table, tj->where, tj->filename);
+				if(use_savepoints && mysql_query(thrconn, "ROLLBACK TO SAVEPOINT mydumper")){
+					g_critical("Rollback to savepoint failed: %s",mysql_error(thrconn));
+				}
 				if(tj->database) g_free(tj->database);
 				if(tj->table) g_free(tj->table);
 				if(tj->where) g_free(tj->where);
@@ -427,6 +476,130 @@ void *process_queue(struct thread_data *td) {
 	return NULL;
 }
 
+void *process_queue_less_locking(struct thread_data *td) {
+	struct configuration *conf= td->conf;
+	// mysql_init is not thread safe, especially in Connector/C
+	g_mutex_lock(init_mutex);
+	MYSQL *thrconn = mysql_init(NULL);
+	g_mutex_unlock(init_mutex);
+	
+	mysql_options(thrconn,MYSQL_READ_DEFAULT_GROUP,"mydumper");
+
+	if (compress_protocol)
+		mysql_options(thrconn,MYSQL_OPT_COMPRESS,NULL);
+
+	if (!mysql_real_connect(thrconn, hostname, username, password, NULL, port, socket_path, 0)) {
+		g_critical("Failed to connect to database: %s", mysql_error(thrconn));
+		exit(EXIT_FAILURE);
+	} else {
+		g_message("Thread %d connected using MySQL connection ID %lu", td->thread_id, mysql_thread_id(thrconn));
+	}
+
+	if ((detected_server == SERVER_TYPE_MYSQL) && mysql_query(thrconn, "SET SESSION wait_timeout = 2147483")){
+		g_warning("Failed to increase wait_timeout: %s", mysql_error(thrconn));
+	}
+	if(!skip_tz && mysql_query(thrconn, "/*!40103 SET TIME_ZONE='+00:00' */")){
+		g_critical("Failed to set time zone: %s",mysql_error(thrconn));
+	}
+	mysql_query(thrconn, "/*!40101 SET NAMES binary*/");
+
+	g_async_queue_push(conf->ready_less_locking,GINT_TO_POINTER(1));
+
+	struct job* job= NULL;
+	struct table_job* tj= NULL;
+	struct tables_job* mj=NULL;
+	struct schema_job* sj= NULL;
+	GList* glj;
+	int first = 1;
+	GString *query= g_string_sized_new(1024);
+	GString *prev_table = g_string_sized_new(100);
+	GString *prev_database = g_string_sized_new(100);
+	
+	for(;;) {
+		GTimeVal tv;
+		g_get_current_time(&tv);
+		g_time_val_add(&tv,1000*1000*1);
+		job=(struct job *)g_async_queue_pop(conf->queue_less_locking);
+		if (shutdown_triggered && (job->type != JOB_SHUTDOWN)) {
+			continue;
+		}
+
+		switch (job->type) {
+			case JOB_LOCK_DUMP_NON_INNODB:
+				mj=(struct tables_job *)job->job_data;
+				glj = g_list_copy(mj->table_job_list);
+				for (glj= g_list_first(glj); glj; glj= g_list_next(glj)) {
+					tj = (struct table_job *)glj->data;
+					if(first){
+						g_string_printf(query, "LOCK TABLES `%s`.`%s` READ LOCAL",tj->database,tj->table);
+						first = 0;
+					}else{
+						if(g_ascii_strcasecmp(prev_database->str, tj->database) || g_ascii_strcasecmp(prev_table->str, tj->table)){
+							g_string_append_printf(query, ", `%s`.`%s` READ LOCAL",tj->database,tj->table);
+						}
+					}
+					g_string_printf(prev_table, "%s", tj->table);
+					g_string_printf(prev_database, "%s", tj->database);
+				}
+				first = 1;
+				if(mysql_query(thrconn,query->str)){
+					g_critical("Non Innodb lock tables fail: %s", mysql_error(thrconn));
+					exit(EXIT_FAILURE);
+				}
+				if (g_atomic_int_dec_and_test(&non_innodb_table_counter) && g_atomic_int_get(&non_innodb_done)) {
+					g_async_queue_push(conf->unlock_tables, GINT_TO_POINTER(1));
+				}
+				for (mj->table_job_list= g_list_first(mj->table_job_list); mj->table_job_list; mj->table_job_list= g_list_next(mj->table_job_list)) {
+					tj = (struct table_job *)mj->table_job_list->data;
+					if (tj->where)
+						g_message("Thread %d dumping data for `%s`.`%s` where %s", td->thread_id, tj->database, tj->table, tj->where);
+					else
+						g_message("Thread %d dumping data for `%s`.`%s`", td->thread_id, tj->database, tj->table);
+					dump_table_data_file(thrconn, tj->database, tj->table, tj->where, tj->filename);
+					if(tj->database) g_free(tj->database);
+					if(tj->table) g_free(tj->table);
+					if(tj->where) g_free(tj->where);
+					if(tj->filename) g_free(tj->filename);
+					g_free(tj);
+				}
+				mysql_query(thrconn, "UNLOCK TABLES /* Non Innodb */");
+				g_free(g_list_first(mj->table_job_list));
+				g_free(mj);
+				g_free(job);
+				break;
+			case JOB_SCHEMA:
+				sj=(struct schema_job *)job->job_data;
+				g_message("Thread %d dumping schema for `%s`.`%s`", td->thread_id, sj->database, sj->table);
+				dump_schema_data(thrconn, sj->database, sj->table, sj->filename);
+				if(sj->database) g_free(sj->database);
+				if(sj->table) g_free(sj->table);
+				if(sj->filename) g_free(sj->filename);
+				g_free(sj);
+				g_free(job);
+				break;
+			case JOB_SHUTDOWN:
+				g_message("Thread %d shutting down", td->thread_id);
+				g_mutex_lock(ll_mutex);
+				less_locking_threads--;
+				g_cond_broadcast(ll_cond);
+				g_mutex_unlock(ll_mutex);
+				if (thrconn)
+					mysql_close(thrconn);
+				g_free(job);
+				mysql_thread_end();
+				return NULL;
+				break;
+			default:
+				g_critical("Something very bad happened!");
+				exit(EXIT_FAILURE);
+		}
+	}
+	if (thrconn)
+		mysql_close(thrconn);
+	mysql_thread_end();
+	return NULL;
+}
+
 MYSQL *reconnect_for_binlog(MYSQL *thrconn) {
 	if (thrconn) {
 		mysql_close(thrconn);
@@ -456,6 +629,8 @@ int main(int argc, char *argv[])
 	g_thread_init(NULL);
 
 	init_mutex = g_mutex_new();
+	ll_mutex = g_mutex_new();
+	ll_cond = g_cond_new();
 
 	context = g_option_context_new("multi-threaded MySQL dumping");
 	GOptionGroup *main_group= g_option_group_new("main", "Main Options", "Main Options", NULL, NULL);
@@ -477,7 +652,24 @@ int main(int argc, char *argv[])
 
 	time_t t;
 	time(&t);localtime_r(&t,&tval);
-
+	
+	//rows chunks have precedence over chunk_filesize 
+	if (rows_per_file > 0 && chunk_filesize > 0){
+		chunk_filesize = 0;
+		g_warning("--chunk-filesize disabled by --rows option");
+	}
+	
+	//until we have an unique option on lock types we need to ensure this
+	if(no_locks)
+		less_locking = 0;
+	
+	/* savepoints workaround to avoid metadata locking issues 
+	   doesnt work for chuncks */
+	if(rows_per_file && use_savepoints){
+		use_savepoints = FALSE;
+		g_warning("--use-savepoints disabled by --rows");
+	}
+	
 	if (!output_directory)
 		output_directory = g_strdup_printf("%s-%04d%02d%02d-%02d%02d%02d",DIRECTORY,
 			tval.tm_year+1900, tval.tm_mon+1, tval.tm_mday,
@@ -554,7 +746,6 @@ int main(int argc, char *argv[])
 	} else {
 		MYSQL *conn= create_main_connection();
 		start_dump(conn);
-		mysql_close(conn);
 	}
 
 	sleep(5);
@@ -581,6 +772,9 @@ MYSQL *create_main_connection()
 		g_critical("Error connecting to database: %s", mysql_error(conn));
 		exit(EXIT_FAILURE);
 	}
+
+	detected_server= detect_server(conn);
+
 	if ((detected_server == SERVER_TYPE_MYSQL) && mysql_query(conn, "SET SESSION wait_timeout = 2147483")){
 		g_warning("Failed to increase wait_timeout: %s", mysql_error(conn));
 	}
@@ -588,7 +782,6 @@ MYSQL *create_main_connection()
 		g_warning("Failed to increase net_write_timeout: %s", mysql_error(conn));
 	}
 
-	detected_server= detect_server(conn);
 	switch (detected_server) {
 		case SERVER_TYPE_MYSQL:
 			g_message("Connected to a MySQL server");
@@ -673,17 +866,30 @@ void *binlog_thread(void *data) {
 
 void start_dump(MYSQL *conn)
 {
-	struct configuration conf = { 1, NULL, NULL, NULL, NULL, 0 };
+	struct configuration conf = { 1, NULL, NULL, NULL, NULL, NULL, NULL, 0 };
 	char *p;
+	char *p2;
+	char *p3;
+	guint64 nits[num_threads];
+	GList* nitl[num_threads];
+	int tn = 0;
+	guint64 min = 0;
 	time_t t;
-
+	struct db_table *dbt;
+	guint n;
+	
+	for(n=0;n<num_threads;n++){
+		nits[n] = 0;
+		nitl[n] = NULL;
+	}
+	
 	if (daemon_mode)
-		p= g_strdup_printf("%s/%d/metadata", output_directory, dump_number);
+		p= g_strdup_printf("%s/%d/metadata.partial", output_directory, dump_number);
 	else
-		p= g_strdup_printf("%s/metadata", output_directory);
+		p= g_strdup_printf("%s/metadata.partial", output_directory);
+	p2 = g_strndup(p, (unsigned)strlen(p)-8);
 
 	FILE* mdfile=g_fopen(p,"w");
-	g_free(p);
 	if(!mdfile) {
 		g_critical("Couldn't write metadata file (%d)",errno);
 		exit(EXIT_FAILURE);
@@ -718,11 +924,11 @@ void start_dump(MYSQL *conn)
 				continue;
 			if (row[tcol] && atoi(row[tcol])>longquery) {
 				if (killqueries) {
-					if (mysql_query(conn,p=g_strdup_printf("KILL %lu",atol(row[icol]))))
+					if (mysql_query(conn,p3=g_strdup_printf("KILL %lu",atol(row[icol]))))
 						g_warning("Could not KILL slow query: %s",mysql_error(conn));
 					else
 						g_warning("Killed a query that was running for %ss",row[tcol]);
-					g_free(p);
+					g_free(p3);
 				} else {
 					g_critical("There are queries in PROCESSLIST running longer than %us, aborting dump,\n\t"
 						"use --long-query-guard to change the guard value, kill queries (--kill-long-queries) or use \n\tdifferent server for dump", longquery);
@@ -741,7 +947,7 @@ void start_dump(MYSQL *conn)
 	} else {
 		g_warning("Executing in no-locks mode, snapshot will notbe consistent");
 	}
-	if (mysql_get_server_version(conn)) {
+	if (mysql_get_server_version(conn) < 40108) {
 		mysql_query(conn, "CREATE TABLE IF NOT EXISTS mysql.mydumperdummy (a INT) ENGINE=INNODB");
 		need_dummy_read=1;
 	}
@@ -766,22 +972,36 @@ void start_dump(MYSQL *conn)
 
 		write_snapshot_info(conn, mdfile);
 	}
+	
+	GThread **threads = g_new(GThread*,num_threads*(less_locking+1));
+	struct thread_data *td= g_new(struct thread_data, num_threads*(less_locking+1));
+	
+	if(less_locking){
+		conf.queue_less_locking = g_async_queue_new();
+		conf.ready_less_locking = g_async_queue_new();
+		less_locking_threads = num_threads;
+		for (n=num_threads; n<num_threads*2; n++) {
+			td[n].conf= &conf;
+			td[n].thread_id= n+1;
+			threads[n] = g_thread_create((GThreadFunc)process_queue_less_locking,&td[n],TRUE,NULL);
+			g_async_queue_pop(conf.ready_less_locking);
+		}
+		g_async_queue_unref(conf.ready_less_locking);
+	}
 
 	conf.queue = g_async_queue_new();
 	conf.ready = g_async_queue_new();
 	conf.unlock_tables= g_async_queue_new();
-
-	guint n;
-	GThread **threads = g_new(GThread*,num_threads);
-	struct thread_data *td= g_new(struct thread_data, num_threads);
+	
 	for (n=0; n<num_threads; n++) {
 		td[n].conf= &conf;
 		td[n].thread_id= n+1;
 		threads[n] = g_thread_create((GThreadFunc)process_queue,&td[n],TRUE,NULL);
 		g_async_queue_pop(conf.ready);
 	}
+	
 	g_async_queue_unref(conf.ready);
-
+	
 	if (db) {
 		dump_database(conn, db);
 	} else {
@@ -800,20 +1020,53 @@ void start_dump(MYSQL *conn)
 		mysql_free_result(databases);
 
 	}
-	struct db_table *dbt;
-	if (!non_innodb_table) {
+	
+	if (!non_innodb_table){
 		g_async_queue_push(conf.unlock_tables, GINT_TO_POINTER(1));
 	}
+	
+	if (less_locking) {
 
-	for (non_innodb_table= g_list_first(non_innodb_table); non_innodb_table; non_innodb_table= g_list_next(non_innodb_table)) {
-		dbt= (struct db_table*) non_innodb_table->data;
-		dump_table(conn, dbt->database, dbt->table, &conf, FALSE);
-		g_atomic_int_inc(&non_innodb_table_counter);
+		for (non_innodb_table= g_list_first(non_innodb_table); non_innodb_table; non_innodb_table= g_list_next(non_innodb_table)) {
+			dbt= (struct db_table*) non_innodb_table->data;
+			tn = 0;
+			min = nits[0];
+			for (n=1; n<num_threads; n++) {
+				if(nits[n] < min){
+					min = nits[n];
+					tn = n;
+				}
+			}
+			nitl[tn]= g_list_append(nitl[tn], dbt);
+			nits[tn] += dbt->datalength;
+		}
+		
+		for (n=0; n<num_threads; n++) {
+			if(nits[n] > 0){
+				g_atomic_int_inc(&non_innodb_table_counter);
+				dump_tables(conn, nitl[n], &conf);
+			}
+		}
+		g_list_free(g_list_first(non_innodb_table));
+
+		
+		g_atomic_int_inc(&non_innodb_done);
+		
+		for (n=0; n<num_threads; n++) {
+			struct job *j = g_new0(struct job,1);
+			j->type = JOB_SHUTDOWN;
+			g_async_queue_push(conf.queue_less_locking,j);
+		}
+	}else{
+		for (non_innodb_table= g_list_first(non_innodb_table); non_innodb_table; non_innodb_table= g_list_next(non_innodb_table)) {
+			dbt= (struct db_table*) non_innodb_table->data;
+			dump_table(conn, dbt->database, dbt->table, &conf, FALSE);
+			g_atomic_int_inc(&non_innodb_table_counter);
+		}
+		g_list_free(g_list_first(non_innodb_table));
+		g_atomic_int_inc(&non_innodb_done);
 	}
-	g_list_free(g_list_first(non_innodb_table));
-
-	g_atomic_int_inc(&non_innodb_done);
-
+	
 	for (innodb_tables= g_list_first(innodb_tables); innodb_tables; innodb_tables= g_list_next(innodb_tables)) {
 		dbt= (struct db_table*) innodb_tables->data;
 		dump_table(conn, dbt->database, dbt->table, &conf, TRUE);
@@ -829,23 +1082,32 @@ void start_dump(MYSQL *conn)
 	}
 	g_list_free(g_list_first(table_schemas));
 
+	if (!no_locks) {
+		g_async_queue_pop(conf.unlock_tables);
+		g_message("Non-InnoDB dump complete, unlocking tables");
+		mysql_query(conn, "UNLOCK TABLES /* FTWRL */");
+	}
+	
 	if (need_binlogs) {
 		get_binlogs(conn, &conf);
 	}
-
-
+	
+	// close main connection 
+	mysql_close(conn);
+	
+	if(less_locking){
+		for (n=num_threads; n<num_threads*2; n++) {
+			g_thread_join(threads[n]);
+		}
+		g_async_queue_unref(conf.queue_less_locking);
+	}
+	
 	for (n=0; n<num_threads; n++) {
 		struct job *j = g_new0(struct job,1);
 		j->type = JOB_SHUTDOWN;
 		g_async_queue_push(conf.queue,j);
 	}
 
-	g_async_queue_pop(conf.unlock_tables);
-	if (!no_locks) {
-		g_message("Non-InnoDB dump complete, unlocking tables");
-		mysql_query(conn, "UNLOCK TABLES");
-	}
-
 	for (n=0; n<num_threads; n++) {
 		g_thread_join(threads[n]);
 	}
@@ -856,6 +1118,9 @@ void start_dump(MYSQL *conn)
 		tval.tm_year+1900, tval.tm_mon+1, tval.tm_mday,
 		tval.tm_hour, tval.tm_min, tval.tm_sec);
 	fclose(mdfile);
+	g_rename(p, p2);
+	g_free(p);
+	g_free(p2);
 	g_message("Finished dump at: %04d-%02d-%02d %02d:%02d:%02d\n",
 		tval.tm_year+1900, tval.tm_mon+1, tval.tm_mday,
 		tval.tm_hour, tval.tm_min, tval.tm_sec);
@@ -954,8 +1219,10 @@ GList * get_chunks_for_table(MYSQL *conn, char *database, char *table, struct co
 			estimated_step = (nmax-nmin)/estimated_chunks+1;
 			cutoff = nmin;
 			while(cutoff<=nmax) {
-				chunks=g_list_append(chunks,g_strdup_printf("%s%s(`%s` >= %llu AND `%s` < %llu)",
+				chunks=g_list_append(chunks,g_strdup_printf("%s%s%s%s(`%s` >= %llu AND `%s` < %llu)",
+						!showed_nulls?"`":"",
 						!showed_nulls?field:"",
+						!showed_nulls?"`":"",
 						!showed_nulls?" IS NULL OR ":"",
 						field, (unsigned long long)cutoff,
 						field, (unsigned long long)(cutoff+estimated_step)));
@@ -1125,6 +1392,12 @@ void dump_database(MYSQL * conn, char *database) {
 		}
 		if (!dump)
 			continue;
+		
+		/* Special tables */
+		if(g_ascii_strcasecmp(database, "mysql") == 0 && (g_ascii_strcasecmp(row[0], "general_log") == 0 || g_ascii_strcasecmp(row[0], "slow_log") == 0)){
+			dump=0;
+			continue;
+		}
 
 		/* Checks PCRE expressions on 'database.table' string */
 		if (regexstring && !check_regex(database,row[0]))
@@ -1134,6 +1407,7 @@ void dump_database(MYSQL * conn, char *database) {
 		struct db_table *dbt = g_new(struct db_table, 1);
 		dbt->database= g_strdup(database);
 		dbt->table= g_strdup(row[0]);
+		dbt->datalength = g_ascii_strtoull(row[6], NULL, 10);
 		if (!g_ascii_strcasecmp("InnoDB", row[ecol])) {
 			innodb_tables= g_list_append(innodb_tables, dbt);
 
@@ -1180,9 +1454,13 @@ void dump_schema_data(MYSQL *conn, char *database, char *table, char *filename)
 
 	query= g_strdup_printf("SHOW CREATE TABLE `%s`.`%s`", database, table);
 	if (mysql_query(conn, query) || !(result= mysql_use_result(conn))) {
-		g_critical("Error dumping schemas (%s.%s): %s", database, table, mysql_error(conn));
+		if(success_on_1146 && mysql_errno(conn) == 1146){
+			g_warning("Error dumping schemas (%s.%s): %s", database, table, mysql_error(conn));
+		}else{
+			g_critical("Error dumping schemas (%s.%s): %s", database, table, mysql_error(conn));
+			errors++;
+		}
 		g_free(query);
-		errors++;
 		return;
 	}
 
@@ -1225,20 +1503,10 @@ void dump_table_data_file(MYSQL *conn, char *database, char *table, char *where,
 		errors++;
 		return;
 	}
-	guint64 rows_count = dump_table_data(conn, (FILE *)outfile, database, table, where);
-	if (!compress_output)
-		fclose((FILE *)outfile);
-	else
-		gzclose((gzFile)outfile);
-
-	if (!rows_count && !build_empty_files) {
-		// dropping the useless file
-		if (remove(filename)) {
- 			g_warning("failed to remove empty file : %s\n", filename);
- 			return;
-		}
-	}
-
+	guint64 rows_count = dump_table_data(conn, (FILE *)outfile, database, table, where, filename);
+	
+	if (!rows_count)
+		g_message("Empty table %s.%s", database,table);
 }
 
 void dump_schema(char *database, char *table, struct configuration *conf) {
@@ -1292,44 +1560,97 @@ void dump_table(MYSQL *conn, char *database, char *table, struct configuration *
 		j->conf=conf;
 		j->type= is_innodb ? JOB_DUMP : JOB_DUMP_NON_INNODB;
 		if (daemon_mode)
-			tj->filename = g_strdup_printf("%s/%d/%s.%s.sql%s", output_directory, dump_number, database, table,(compress_output?".gz":""));
+			tj->filename = g_strdup_printf("%s/%d/%s.%s%s.sql%s", output_directory, dump_number, database, table,(chunk_filesize?".00001":""),(compress_output?".gz":""));
 		else
-			tj->filename = g_strdup_printf("%s/%s.%s.sql%s", output_directory, database, table,(compress_output?".gz":""));
+			tj->filename = g_strdup_printf("%s/%s.%s%s.sql%s", output_directory, database, table,(chunk_filesize?".00001":""),(compress_output?".gz":""));
 		g_async_queue_push(conf->queue,j);
 		return;
 	}
 }
 
+void dump_tables(MYSQL *conn, GList *noninnodb_tables_list, struct configuration *conf){
+	struct db_table* dbt;
+	GList * chunks = NULL;
+
+	struct job *j = g_new0(struct job,1);
+	struct tables_job *tjs = g_new0(struct tables_job,1);
+	j->conf=conf;
+	j->type=JOB_LOCK_DUMP_NON_INNODB;
+	j->job_data=(void*) tjs;
+
+	for (noninnodb_tables_list= g_list_first(noninnodb_tables_list); noninnodb_tables_list; noninnodb_tables_list= g_list_next(noninnodb_tables_list)) {
+		dbt = (struct db_table*) noninnodb_tables_list->data;
+
+		if (rows_per_file)
+			chunks = get_chunks_for_table(conn, dbt->database, dbt->table, conf);
+
+		if(chunks){
+			int nchunk=0;
+			for (chunks = g_list_first(chunks); chunks; chunks=g_list_next(chunks)) {
+				struct table_job *tj = g_new0(struct table_job,1);
+				tj->database = g_strdup_printf("%s",dbt->database);
+				tj->table = g_strdup_printf("%s",dbt->table);
+				if (daemon_mode)
+					tj->filename=g_strdup_printf("%s/%d/%s.%s.%05d.sql%s", output_directory, dump_number, dbt->database, dbt->table, nchunk,(compress_output?".gz":""));
+				else
+					tj->filename=g_strdup_printf("%s/%s.%s.%05d.sql%s", output_directory, dbt->database, dbt->table, nchunk,(compress_output?".gz":""));
+				tj->where=(char *)chunks->data;
+				tjs->table_job_list= g_list_append(tjs->table_job_list, tj);
+				nchunk++;
+			}
+		}else{
+			struct table_job *tj = g_new0(struct table_job,1);
+			tj->database = g_strdup_printf("%s",dbt->database);
+			tj->table = g_strdup_printf("%s",dbt->table);
+			if (daemon_mode)
+				tj->filename = g_strdup_printf("%s/%d/%s.%s%s.sql%s", output_directory, dump_number, dbt->database, dbt->table,(chunk_filesize?".00001":""),(compress_output?".gz":""));
+			else
+				tj->filename = g_strdup_printf("%s/%s.%s%s.sql%s", output_directory, dbt->database, dbt->table,(chunk_filesize?".00001":""),(compress_output?".gz":""));
+			tj->where = NULL;
+			tjs->table_job_list= g_list_append(tjs->table_job_list, tj);
+		}
+	}
+	g_async_queue_push(conf->queue_less_locking,j);
+}
+
 /* Do actual data chunk reading/writing magic */
-guint64 dump_table_data(MYSQL * conn, FILE *file, char *database, char *table, char *where)
+guint64 dump_table_data(MYSQL * conn, FILE *file, char *database, char *table, char *where, char *filename)
 {
 	guint i;
+	guint fn = 1;
+	guint st_in_file = 0;
 	guint num_fields = 0;
 	guint64 num_rows = 0;
+	guint64 num_rows_st = 0;
 	MYSQL_RES *result = NULL;
 	char *query = NULL;
+	gchar *fcfile = NULL;
+	gchar* filename_prefix = NULL;
+	
+	fcfile = g_strdup (filename);
+	
+	if(chunk_filesize){
+		gchar** split_filename= g_strsplit(filename, ".00001.sql", 0);
+		filename_prefix= split_filename[0];
+		g_free(split_filename);
+	}
 
+	
 	/* Ghm, not sure if this should be statement_size - but default isn't too big for now */
 	GString* statement = g_string_sized_new(statement_size);
-
-	if (detected_server == SERVER_TYPE_MYSQL) {
-		g_string_printf(statement,"/*!40101 SET NAMES binary*/;\n");
-		g_string_append(statement,"/*!40014 SET FOREIGN_KEY_CHECKS=0*/;\n");
-	} else {
-		g_string_printf(statement,"SET FOREIGN_KEY_CHECKS=0;\n");
-	}
-
-	if (!write_data(file,statement)) {
-		g_critical("Could not write out data for %s.%s", database, table);
-		return num_rows;
-	}
-
+	GString* statement_row = g_string_sized_new(0);
+	
 	/* Poor man's database code */
  	query = g_strdup_printf("SELECT %s * FROM `%s`.`%s` %s %s", (detected_server == SERVER_TYPE_MYSQL) ? "/*!40001 SQL_NO_CACHE */" : "", database, table, where?"WHERE":"",where?where:"");
 	if (mysql_query(conn, query) || !(result=mysql_use_result(conn))) {
-		g_critical("Error dumping table (%s.%s) data: %s ",database, table, mysql_error(conn));
+		//ERROR 1146 
+		if(success_on_1146 && mysql_errno(conn) == 1146){
+			g_warning("Error dumping table (%s.%s) data: %s ",database, table, mysql_error(conn));
+		}else{
+			g_critical("Error dumping table (%s.%s) data: %s ",database, table, mysql_error(conn));
+			errors++;
+		}
 		g_free(query);
-		errors++;
 		return num_rows;
 	}
 
@@ -1348,38 +1669,87 @@ guint64 dump_table_data(MYSQL * conn, FILE *file, char *database, char *table, c
 		gulong *lengths = mysql_fetch_lengths(result);
 		num_rows++;
 
-		if (!statement->len)
-			g_string_printf(statement, "INSERT INTO `%s` VALUES\n(", table);
-		else
-			g_string_append(statement, ",\n(");
+		if (!statement->len){
+			if(!st_in_file){
+				if (detected_server == SERVER_TYPE_MYSQL) {
+					g_string_printf(statement,"/*!40101 SET NAMES binary*/;\n");
+					g_string_append(statement,"/*!40014 SET FOREIGN_KEY_CHECKS=0*/;\n");
+					if (!skip_tz) {
+					  g_string_append(statement,"/*!40103 SET TIME_ZONE='+00:00' */;\n");
+					}
+				} else {
+					g_string_printf(statement,"SET FOREIGN_KEY_CHECKS=0;\n");
+				}
+
+				if (!write_data(file,statement)) {
+					g_critical("Could not write out data for %s.%s", database, table);
+					return num_rows;
+				}
+			}
+			g_string_printf(statement, "INSERT INTO `%s` VALUES", table);
+			num_rows_st = 0;
+		}
+		
+		if (statement_row->len) {
+			g_string_append(statement, statement_row->str);
+			g_string_set_size(statement_row,0);
+			num_rows_st++;
+		}
+		
+		g_string_append(statement_row, "\n(");
 
 		for (i = 0; i < num_fields; i++) {
 			/* Don't escape safe formats, saves some time */
 			if (!row[i]) {
-				g_string_append(statement, "NULL");
+				g_string_append(statement_row, "NULL");
 			} else if (fields[i].flags & NUM_FLAG) {
-				g_string_append(statement, row[i]);
+				g_string_append(statement_row, row[i]);
 			} else {
 				/* We reuse buffers for string escaping, growing is expensive just at the beginning */
 				g_string_set_size(escaped, lengths[i]*2+1);
 				mysql_real_escape_string(conn, escaped->str, row[i], lengths[i]);
-				g_string_append_c(statement,'\"');
-				g_string_append(statement,escaped->str);
-				g_string_append_c(statement,'\"');
+				g_string_append_c(statement_row,'\"');
+				g_string_append(statement_row,escaped->str);
+				g_string_append_c(statement_row,'\"');
 			}
 			if (i < num_fields - 1) {
-				g_string_append_c(statement,',');
+				g_string_append_c(statement_row,',');
 			} else {
-				/* INSERT statement is closed once over limit */
-				if (statement->len > statement_size) {
-					g_string_append(statement,");\n");
+				g_string_append_c(statement_row,')');
+				/* INSERT statement is closed before over limit */
+				if(statement->len+statement_row->len+1 > statement_size) {
+					if(num_rows_st == 0){
+						g_string_append(statement, statement_row->str);
+						g_string_set_size(statement_row,0);
+						g_warning("Row bigger than statement_size for %s.%s", database, table);
+					}
+					g_string_append(statement,";\n");
+
 					if (!write_data(file,statement)) {
 						g_critical("Could not write out data for %s.%s", database, table);
 						goto cleanup;
+					}else{
+						st_in_file++;
+						if(chunk_filesize && st_in_file*(guint)ceil((float)statement_size/1024/1024) > chunk_filesize){
+							fn++;
+							fcfile = g_strdup_printf("%s.%05d.sql%s", filename_prefix,fn,(compress_output?".gz":""));
+							if (!compress_output){
+								fclose((FILE *)file);
+								file = g_fopen(fcfile, "w");
+                            } else {
+								gzclose((gzFile)file);
+                                file = (void*) gzopen(fcfile, "w");
+                            }
+							st_in_file = 0;
+						}
 					}
 					g_string_set_size(statement,0);
 				} else {
-					g_string_append_c(statement,')');
+					if(num_rows_st)
+						g_string_append_c(statement,',');
+					g_string_append(statement, statement_row->str);
+					num_rows_st++;
+					g_string_set_size(statement_row,0);
 				}
 			}
 		}
@@ -1389,15 +1759,12 @@ guint64 dump_table_data(MYSQL * conn, FILE *file, char *database, char *table, c
 	}
 
 	if (statement->len > 0) {
-		if (!write_data(file,statement)) {
-			g_critical("Could not write out data for %s.%s", database, table);
-			goto cleanup;
-		}
-		g_string_printf(statement,";\n");
+		g_string_append(statement,";\n");
 		if (!write_data(file,statement)) {
 			g_critical("Could not write out closing newline for %s.%s, now this is sad!", database, table);
 			goto cleanup;
 		}
+		st_in_file++;
 	}
 
 cleanup:
@@ -1409,7 +1776,26 @@ cleanup:
 	if (result) {
 		mysql_free_result(result);
 	}
-
+	
+	if (!compress_output){
+		fclose((FILE *)file);
+	} else {
+		gzclose((gzFile)file);
+	}
+	
+	if (!st_in_file && !build_empty_files) {
+		// dropping the useless file
+		if (remove(fcfile)) {
+ 			g_warning("Failed to remove empty file : %s\n", fcfile);
+		}
+	}else if(chunk_filesize && fn == 1){
+		fcfile = g_strdup_printf("%s.sql%s", filename_prefix,(compress_output?".gz":""));
+		g_rename(filename, fcfile);
+	}
+	
+	g_free(filename_prefix);
+	g_free(fcfile);
+	
 	return num_rows;
 }
 
